/*
 *
 * Copyright  2010-2013 Balzs Tth <tbalazs@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <iostream>
#include <fstream>
#include <cmath>
#include <CL/opencl.h>

#include <GL/glew.h>
#include <GL/freeglut.h>

#include "clwrapper.hpp"

ClWrapper cl;

// *****************
// OpenCL processing
// *****************
cl_program simulationProgram;

int width = 512;
int height = 512;

size_t visualizationBufferSize[2];
cl_float4* visualizationBufferCPU;
cl_mem visualizationBufferGPU;

//kernels
cl_kernel visualizationClearKernel;
cl_kernel visualizationKernel;


void resetSimulation();

const size_t bodyNum = 100;
// position + velocity
cl_mem particlesBuffer;

void initSimulation(){
  simulationProgram = cl.createProgram("../kernels/programs.cl");

  visualizationBufferSize[0] = width;
  visualizationBufferSize[1] = height;
  visualizationBufferCPU = new cl_float4[width * height];
  visualizationBufferGPU = clCreateBuffer(cl.context(), CL_MEM_READ_WRITE, sizeof(cl_float4) * width * height, NULL, NULL);
  visualizationKernel = cl.createKernel(simulationProgram, "visualization");
  visualizationClearKernel = cl.createKernel(simulationProgram, "visualizationClear");

  //TODO create simulation kernel
  particlesBuffer = clCreateBuffer(cl.context(), CL_MEM_READ_WRITE, sizeof(cl_float4) * bodyNum, NULL, NULL);

  resetSimulation();
}

void resetSimulation(){
	//TODO
	// fill particlesbuffer with random positions and velocities
	//
}

void simulationStep(){
	//TODO
	// make one simulation step, which updates particlesBuffer
}

void visualizationStep(){
  CL_SAFE_CALL( clSetKernelArg(visualizationClearKernel, 0, sizeof(int), &width));
  CL_SAFE_CALL( clSetKernelArg(visualizationClearKernel, 1, sizeof(int), &height));
  CL_SAFE_CALL( clSetKernelArg(visualizationClearKernel, 2, sizeof(cl_mem), &visualizationBufferGPU));
  CL_SAFE_CALL( clEnqueueNDRangeKernel(cl.cqueue(), visualizationClearKernel,
				       2, NULL, visualizationBufferSize, NULL,
				       0, NULL, NULL));
  clFinish(cl.cqueue());

  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 0, sizeof(int), &width) );
  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 1, sizeof(int), &height) );
  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 2, sizeof(cl_mem), &visualizationBufferGPU) );
  CL_SAFE_CALL( clSetKernelArg(visualizationKernel, 3, sizeof(cl_mem), &particlesBuffer));

  CL_SAFE_CALL( clEnqueueNDRangeKernel(cl.cqueue(), visualizationKernel,
				       1, NULL, &bodyNum, NULL,
				       0, NULL, NULL) );
  clFinish(cl.cqueue());

  CL_SAFE_CALL( clEnqueueReadBuffer(cl.cqueue(), visualizationBufferGPU, CL_TRUE,
				    0, sizeof(cl_float4) * width * height, visualizationBufferCPU,
				    0, NULL, NULL) );
  glDrawPixels(width, height, GL_RGBA, GL_FLOAT, visualizationBufferCPU);
}

// OpenGL
bool keysPressed[256];

void initOpenGL(){
  glewExperimental = GL_TRUE;
  GLenum err = glewInit();
  if (GLEW_OK != err)     {
    std::cerr << "Error: "<< glewGetErrorString(err) << std::endl;
  } else {
    if (GLEW_VERSION_3_0)
      {
	std::cout<< "Driver supports OpenGL 3.0\nDetails:"<< std::endl;
	std::cout << "  Using GLEW " << glewGetString(GLEW_VERSION)<< std::endl;
	std::cout << "  Vendor: " << glGetString (GL_VENDOR)<< std::endl;
	std::cout << "  Renderer: " << glGetString (GL_RENDERER)<< std::endl;
	std::cout << "  Version: " << glGetString (GL_VERSION)<< std::endl;
	std::cout << "  GLSL: " << glGetString (GL_SHADING_LANGUAGE_VERSION)<< std::endl;
      }
  }

  glClearColor(0.17f, 0.4f, 0.6f, 1.0f);
}

void display(){
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glDisable(GL_DEPTH_TEST);

  simulationStep();
  visualizationStep();

  glEnable(GL_DEPTH_TEST);
  glutSwapBuffers();
}

void idle(){
  glutPostRedisplay();
}

void keyDown(unsigned char key, int x, int y){
  keysPressed[key] = true;
}

void keyUp(unsigned char key, int x, int y){
  keysPressed[key] = false;
  switch(key){

  case 'r':
    resetSimulation();
    break;

  case 27:
    exit(0);
    break;
  }
}

void mouseClick(int button, int state, int x, int y){

}

void mouseMove(int x, int y){

}

void reshape(int newWidth, int newHeight){
        width = newWidth;
        height = newHeight;
        glViewport(0, 0, width, height);
}

int main(int argc, char* argv[]){

  glutInit(&argc, argv);
  glutInitContextVersion (3, 0);
  glutInitContextFlags (GLUT_CORE_PROFILE | GLUT_DEBUG);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
  glutInitWindowSize(width, height);
  glutCreateWindow("NBODY");

  initOpenGL();

  glutDisplayFunc(display);
  glutIdleFunc(idle);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyDown);
  glutKeyboardUpFunc(keyUp);
  glutMouseFunc(mouseClick);
  glutMotionFunc(mouseMove);

  // OpenCL processing
  initSimulation();

  glutMainLoop();
  return(0);
}
